@@ -1,5 +1,5 @@
 /* -*- C -*-
- * $Id: cptr.c 44754 2014-01-30 03:49:07Z usa $
+ * $Id: cptr.c 44903 2014-02-10 11:45:14Z naruse $
  */
 
 #include <ruby/ruby.h>
@@ -7,15 +7,6 @@
 #include <ctype.h>
 #include "dl.h"
 
-#ifdef PRIsVALUE
-# define RB_OBJ_CLASSNAME(obj) rb_obj_class(obj)
-# define RB_OBJ_STRING(obj) (obj)
-#else
-# define PRIsVALUE "s"
-# define RB_OBJ_CLASSNAME(obj) rb_obj_classname(obj)
-# define RB_OBJ_STRING(obj) StringValueCStr(obj)
-#endif
-
 VALUE rb_cDLCPtr;
 
 static inline freefunc_t
@@ -72,27 +63,17 @@ static const rb_data_type_t dlptr_data_type = {
     {dlptr_mark, dlptr_free, dlptr_memsize,},
 };
 
-void
-dlptr_init(VALUE val)
-{
-    struct ptr_data *data;
-
-    TypedData_Get_Struct(val, struct ptr_data, &dlptr_data_type, data);
-    OBJ_TAINT(val);
-}
-
 VALUE
 rb_dlptr_new2(VALUE klass, void *ptr, long size, freefunc_t func)
 {
     struct ptr_data *data;
     VALUE val;
 
-    rb_secure(4);
     val = TypedData_Make_Struct(klass, struct ptr_data, &dlptr_data_type, data);
     data->ptr = ptr;
     data->free = func;
     data->size = size;
-    dlptr_init(val);
+    OBJ_TAINT(val);
 
     return val;
 }
@@ -108,7 +89,6 @@ rb_dlptr_malloc(long size, freefunc_t func)
 {
     void *ptr;
 
-    rb_secure(4);
     ptr = ruby_xmalloc((size_t)size);
     memset(ptr,0,(size_t)size);
     return rb_dlptr_new(ptr, size, func);
@@ -140,7 +120,6 @@ rb_dlptr_s_allocate(VALUE klass)
     VALUE obj;
     struct ptr_data *data;
 
-    rb_secure(4);
     obj = TypedData_Make_Struct(klass, struct ptr_data, &dlptr_data_type, data);
     data->ptr = 0;
     data->size = 0;
@@ -409,10 +388,11 @@ static VALUE
 rb_dlptr_inspect(VALUE self)
 {
     struct ptr_data *data;
+    char str[1024];
 
     TypedData_Get_Struct(self, struct ptr_data, &dlptr_data_type, data);
     return rb_sprintf("#<%"PRIsVALUE":%p ptr=%p size=%ld free=%p>",
-		      RB_OBJ_CLASSNAME(self), data, data->ptr, data->size, data->free);
+		      rb_obj_class(self), data, data->ptr, data->size, data->free);
 }
 
 /*
